#!/usr/bin/perl

# NDMP Index File Parser
# Takes a backup number and an NDMP index file as arguments and adds the files
# and directories therein to the FILES_TABLE and DIRS_TABLE, respectively. If
# no index file is provided, an attempt will be made to mount the backup,
# extract the index file, and proceed.
#
# The verbose (-v) and debug (-d) options send their output to STDERR. If
# calling this script from elsewhere, consider using the -v option and
# redirecting STDERR to append to <LOG_DIR>/ndmp_index_parser.log.
#
# This script returns 0 on success and the error code returned by the DBD::Pg
# module if an error occurs. If any file names were not able to be placed in
# the database (typically due to character encoding problems), the script
# returns 254. This is not, strictly speaking, a failure, and this exit code
# should be checked for when calling.
#
# NOTE CentOS 5 and CentOS 6 have different versions of the DBD::Pg module. The
# CentOS 5 version is heavily deprecated and requires slightly different usage
# than the CentOS 6 version. The NEW_DBD_VERS constant is used to differentiate
# between the two usages.

use strict;
use warnings;
use DBI;
use DBD::Pg;
use URI::Escape;
use Getopt::Std;
use POSIX;

use constant {
	BPUTIL => "/usr/bp/bin/bputil",
	MASTER_INI => "/usr/bp/bpinit/master.ini",
	DATABASE => "bpdb",
	DB_USER => "postgres",
	DIRS_TABLE => "bp.dirs_non_ctar",
	FILES_TABLE => "bp.backup_files_non_ctar",
	END_BACKUP_NO => 2147483647,    # Convention for non-chained backups
	FT_REG => 0,                    # Internal file type identifier for regular files
	FT_DIR => 53,                   # Internal file type identifier for directories
	FT_CONTAINER => 57,             # Internal file type identifier for containers
	NDMP_MOUNT_PATH => "/tmp/ndmp", # Temporary location to mount NDMP backups
	VOLUME => '@@@:',               # Special volume container entry
	NEW_DBD_VERS => 2.0,            # Earliest known version to use new functions
	VERSION => "1.4.2",             # Version of this script
};

# Constants to provide functionality similar to an associative array for the
# lookup table. Using an actual associative array requires 6 additional hashes
# per entry, whereas a standard array only requires 6 additional ints. Since
# Perl's hashes are 120 bytes each, and its ints are 8 bytes each, the overhead
# of an associative array would be 720 bytes, compared to just 48 bytes with a
# standard array. This should reduce memory usage substantially with large file
# sets (672 bytes per file).

use constant {
	NAME => 0,
	PARENT => 1,
	MOD_DATE => 2,
	BLOCK_OFFSET => 3,
	TYPE => 4,
	DIR_NO => 5,
	SIZE => 6,
};

our (%opts, $backup_no, $index_file, %lookup, $mounted);

my $dbh;   # Database handler
my $key;   # Key for hash-map
my $query; # Query string
my $sth;   # Statement handler
my $sth2;  # Statement handler
my $LOG_LEVEL = 0; # 0 = Quiet, 1 = Verbose, 2 = Debug

# Print usage and exit
sub HELP_MESSAGE {
	print STDERR "This script parses NDMP index files and adds the names ",
		"and file attributes of files and directories in the backup to ",
		"a database.\n",
		"Usage: $0 [-h] [-l <LOG_LEVEL>] -b <BACKUP_NUMBER> [-f <FILE>]\n",
		"-h\tPrint this help and exit.\n",
		"-l LOG_LEVEL\t0 = Quiet (errors and warnings only), 1 = Verbose, 2 = Debug.\n",
		"-b NUM\tBackup number.\n",
		"-f FILE\tNDMP index file. If not provided, the file will be inferred ",
		"based on the backup number.\n";
	exit;
}

$Getopt::Std::STANDARD_HELP_VERSION = 1;
sub VERSION_MESSAGE {
	print STDERR "v" . VERSION . "\n";
}

# Return a timestamp for the current time as a string
sub timestamp {
	return strftime "%F %T", localtime;
}

# Prints a log message to STDERR including a timestamp. Takes two arguments,
# the log level and the log message. The message will only be printed if the
# $LOG_LEVEL is greater than or equal to the provided log level.
# logging(0, "Message") will always print, but logging(1, "Message") will only
# print if $LOG_LEVEL is set to 1 or higher.
sub logging {
	if (defined $_[0] and defined $_[1] and $LOG_LEVEL >= $_[0]) {
		print STDERR "[" . timestamp . "] " . $_[1] . "\n";
	}
}

# Connect to the database
sub db_connect {
	logging(1, "Connecting to database...");
	$dbh = DBI->connect("dbi:Pg:dbname=" . DATABASE, DB_USER, "")
			or die "Could not connect to database server\n";
}

# Disconnect from the database
sub db_disconnect {
	logging(1, "Disconnect from database\n");
	$dbh->disconnect();
}

# Mount backup
sub mount {
	logging(1, "Mount backup $backup_no");
	system("/usr/bp/bin/cmc_dump_mount", "--path=" . NDMP_MOUNT_PATH, "mount",
			$backup_no) == 0 or die "Failed to mount backup $backup_no\n";
	$index_file = NDMP_MOUNT_PATH . "/_dump/$backup_no/index";
	$mounted = 1;
}

# Unmount backup if $mounted
sub unmount {
	if ($mounted) {
		logging(1, "Unmount backup $backup_no");
		system("/usr/bp/bin/cmc_dump_mount", "--path=" . NDMP_MOUNT_PATH, "destroy",
				$backup_no) == 0 or logging(0, "Failed to unmount backup $backup_no");
	}
}

# Prepare to exit, especially in case of an error. Call before dying.
sub cleanup {
	&db_disconnect;
	&unmount;
}

# Initialize the script (read options, connect to database)
sub init {
	my $opt_string = 'hl:f:b:';
	getopts("$opt_string", \%opts) or HELP_MESSAGE();
	HELP_MESSAGE() if $opts{h} or not %opts; # exits
	$index_file = $opts{f};
	$backup_no = $opts{b};
	# Set LOG_LEVEL from command line or master.ini [NDMP] ParserLogLevel, preferring command line
	if (defined $opts{l}) {
		$LOG_LEVEL = $opts{l};
	} else {
		open my $conf, '-|', BPUTIL . " -g NDMP ParserLogLevel $LOG_LEVEL " . MASTER_INI;
		$LOG_LEVEL = <$conf>;
		close $conf;
	}
	# Ensure LOG_LEVEL is at least 0
	if ($LOG_LEVEL < 0) {$LOG_LEVEL = 0};
	$mounted = 0;
	&db_connect;
}

# Get the full filepath from a node number
sub get_path_from_node {
	my $node = $_[0];
	my $path = "";

	while ($node and $node != 2) { # The root node is always 2
		$path = "/$lookup{$node}[NAME]" . $path;
		$node = $lookup{$node}[PARENT];
	}

	return $path;
}

&init;

# Unless an index file was specified, and exists, attempt to mount the backup
# and use the index file contained within
&mount unless ($opts{f} and -e $opts{f});

logging(1, "Index File: $index_file, Backup No.: $backup_no");

# Prepare INSERT/SELECT statements for DIRS_TABLE
$sth = $dbh->prepare("INSERT INTO " . DIRS_TABLE . " (dname) SELECT ? " .
	"WHERE NOT EXISTS (SELECT TRUE FROM " . DIRS_TABLE . " WHERE dname = ?)");
$sth2 = $dbh->prepare("SELECT dir_no FROM " . DIRS_TABLE .  " WHERE dname = ?");

# The blank directory entry should be added by the schema definition, but is
# included here for safety and consistencty.
$sth->execute('', '');
$sth2->execute('');
my $volume_dir_no = $sth2->fetchrow_array();

# Give cmc_dump_mount time to finish mounting. Using sleep is not ideal, but
# all the implicit forking and asynchronous commands means waiting for signals
# can potentially hang indefinitely; this approach allows us to set a timeout.
my $time = 1800;
while ($mounted and not -e $index_file and $time) {
	if ($time % 10 == 0) {
		logging(1, "Mounting backup $backup_no, $time seconds remaining");
	}
	$time -= 1;
	sleep 1;
}

if (open(NDMP_INDEX, "<", $index_file)) {
	logging(1, "Parse $index_file...");

	# Read through the index file and add each entry to the lookup hash table.
	while (<NDMP_INDEX>) {
		chomp;
		# Type - Parent - Name - Host Type? - Node
		if ($_ =~ /^DHd (\d+) (.+) UNIX (\d+)/) {
			# Node
			if ($2 ne '.' and $2 ne '..') {
				$lookup{$3}[NAME] = uri_unescape($2);
				$lookup{$3}[PARENT] = $1;
				$lookup{$3}[MOD_DATE] = -1;
				$lookup{$3}[BLOCK_OFFSET] = 0;
				$lookup{$3}[TYPE] = FT_REG;
				$lookup{$3}[SIZE] = 0;
			}
		} elsif ($_ =~ /^DHn (\d+) UNIX fd m(\d+) u(\d+) g(\d+) tm(\d+) @(\d+)/) {
			# Directory
			if (exists $lookup{$1}) {
				$lookup{$1}[MOD_DATE] = $5;
				$lookup{$1}[BLOCK_OFFSET] = $6;
				$lookup{$1}[TYPE] = FT_DIR;
				$lookup{$3}[DIR_NO] = $volume_dir_no;
			}
		} elsif ($_ =~ /^DHn (\d+) UNIX f- m(\d+) u(\d+) g(\d+) s(\d+) tm(\d+) @(\d+)/) {
			# File
			if (exists $lookup{$1}) {
				$lookup{$1}[SIZE] = $5;
				$lookup{$1}[MOD_DATE] = $6;
				$lookup{$1}[BLOCK_OFFSET] = $7;
			}
		} elsif ($_ =~ /^DHr (\d+)$/) {
			# Root
			unless (exists $lookup{$1}) {
				$lookup{$1}[NAME] = '/';
				$lookup{$1}[PARENT] = 0;
				$lookup{$1}[MOD_DATE] = 0;
				$lookup{$1}[BLOCK_OFFSET] = 0;
				$lookup{$1}[TYPE] = FT_DIR;
			}
		} else {
			# No matches
			logging(2, "$_");
		}
	}

	logging(1, "Done parsing");
	close(NDMP_INDEX);
} else {
	&cleanup;
	die "Cannot open '$index_file': $!\n";
}

# Parsing finished, no need to leave the backup mounted
&unmount;

# Add all directory entries to DIRS_TABLE
logging(1, "Insert directories...");
#$dbh->begin_work();

foreach $key (keys %lookup) {
	# Only run on directories
	next if (not exists($lookup{$key}[TYPE]) or $lookup{$key}[TYPE] != FT_DIR);

	my $path = VOLUME . get_path_from_node($key) . "/";

	# Add directory if unique
	$sth->execute($path, $path);
	if ($sth->err) {
		logging(0, "ERROR: Failed to add directory $path");
	}

	# Get DIR_NO to update the current item in the lookup table
	$sth2->execute($path);
	if ($sth2->err) {
		logging(0, "ERROR: Could not retrieve DIR_NO for $path");
		# This will make the entry unusable, remove it from the lookup table
		delete $lookup{$key};
	} else {
		$lookup{$key}[DIR_NO] = $sth2->fetchrow_array();
	}
}

#$dbh->commit();

if ($dbh->err) {
	logging(0, "Failed to commit directories to database.");
	&db_disconnect;
	exit $dbh->err;
}

logging(1, "Done inserting directories");

$sth = undef;
$sth2 = undef;

# Generate tuples to COPY into FILES_TABLE
logging(1, "Insert files...");
$dbh->do("COPY " . FILES_TABLE . "(backup_no, end_backup_no, dir_no, fname,
	type, mod_date, blk_offset, size, blocks) FROM STDIN");

# Add "@@@:" entry manually
if ($DBD::Pg::VERSION < NEW_DBD_VERS) {
	$dbh->pg_putline(
		"$backup_no\t" .
		END_BACKUP_NO . "\t" .
		"$volume_dir_no\t" .
		VOLUME . "\t" .
		FT_CONTAINER . "\t" .
		"0\t" .
		"0\t" .
		"0\t" .
		"0\n");
} else {
	$dbh->pg_putcopydata(
		"$backup_no\t" .
		END_BACKUP_NO . "\t" .
		"$volume_dir_no\t" .
		VOLUME . "\t" .
		FT_CONTAINER . "\t" .
		"0\t" .
		"0\t" .
		"0\t" .
		"0\n");
}

# Add all other entries
foreach $key (keys %lookup) {
	# Some items may need to be skipped (such as unmodified files from incrementals)
	if ($key == 0
			or not exists($lookup{$key}[PARENT])
			or $lookup{$key}[PARENT] == 0
			or $lookup{$key}[MOD_DATE] == -1) {
		next;
	}

	if ($DBD::Pg::VERSION < NEW_DBD_VERS) {
		$dbh->pg_putline(
			"$backup_no\t" .
			END_BACKUP_NO . "\t" .
			"$lookup{$lookup{$key}[PARENT]}[DIR_NO]\t" .
			"$lookup{$key}[NAME]\t" .
			"$lookup{$key}[TYPE]\t" .
			"$lookup{$key}[MOD_DATE]\t" .
			"$lookup{$key}[BLOCK_OFFSET]\t" .
			"$lookup{$key}[SIZE]\t" .
			"0\n");
	} else {
		$dbh->pg_putcopydata(
			"$backup_no\t" .
			END_BACKUP_NO . "\t" .
			"$lookup{$lookup{$key}[PARENT]}[DIR_NO]\t" .
			"$lookup{$key}[NAME]\t" .
			"$lookup{$key}[TYPE]\t" .
			"$lookup{$key}[MOD_DATE]\t" .
			"$lookup{$key}[BLOCK_OFFSET]\t" .
			"$lookup{$key}[SIZE]\t" .
			"0\n");
	}
}

logging(1, "Done inserting files");

if ($DBD::Pg::VERSION < NEW_DBD_VERS) {
	$dbh->pg_endcopy();
} else {
	$dbh->pg_putcopyend();
}

# If there was an error exiting COPY mode, the transaction will fail to commit;
# this results in a backup which has no file names in the database, making
# selective restores impossible. Attempt to insert the files one at a time
# using prepared statements to get as many file names as possible into the
# database and warn the user that some files may be unavailable for selective
# restore. Note that this is MUCH (~60 times) slower and should be used only as
# a last resort.
if ($dbh->err) {
	logging(0, "WARNING: Backup $backup_no (or some files in it) may be ",
		"ineligible for selective restore.");

	# The CentOS 5 version of the DBD::Pg module does not reset the internal
	# COPY flag if pg_endcopy() fails. The database connection needs to be
	# reset to issue further commands.
	if ($DBD::Pg::VERSION < NEW_DBD_VERS) {
		&db_disconnect;
		&db_connect;
	}

	$sth = $dbh->prepare("INSERT INTO " . FILES_TABLE . " (backup_no,
		end_backup_no, dir_no, fname, type, mod_date, blk_offset, size, blocks)
		VALUES(?, ?, ?, ?, ?, ?, ?, ?, 0)");

	# Add "@@@:" entry manually
	$sth->execute($backup_no, END_BACKUP_NO, $volume_dir_no, VOLUME, FT_CONTAINER, 0, 0, 0);

	foreach $key (keys %lookup) {
		next if ($key == 0 or not exists($lookup{$key}[PARENT]) or $lookup{$key}[PARENT] == 0
				or not exists($lookup{$lookup{$key}[PARENT]}[DIR_NO]));

		$sth->execute($backup_no, END_BACKUP_NO,
			$lookup{$lookup{$key}[PARENT]}[DIR_NO], $lookup{$key}[NAME],
			$lookup{$key}[TYPE], $lookup{$key}[MOD_DATE],
			$lookup{$key}[BLOCK_OFFSET], $lookup{$key}[SIZE]);

		if ($sth->err and $sth->err == 7) {
			# A return value of 7 indicates a fatal error was returned, the
			# last query failed. This is probably due to a non-UTF-8 filename,
			# so let's log the filename.
			my $bad_path = get_path_from_node($key);
			logging(0, "WARNING: Failed to add file $bad_path");
		}
	}

	&db_disconnect;
	exit 254; # Let caller know something went wrong and to flag this backup with a warning.
}

&db_disconnect;
